unit util;

interface

uses windows;

const
  SECURITY_NT_AUTHORITY: TSIDIdentifierAuthority =
    (Value: (0, 0, 0, 0, 0, 5));
  SECURITY_BUILTIN_DOMAIN_RID = $00000020;
  DOMAIN_ALIAS_RID_ADMINS = $00000220;

function IsAdmin: Boolean;

function setWMIAccessRights(user:string):string;

{$EXTERNALSYM ConvertSecurityDescriptorToStringSecurityDescriptor}
function ConvertSecurityDescriptorToStringSecurityDescriptor(
             SecurityDescriptor : PSECURITY_DESCRIPTOR;
             RequestedStringSDRevision : DWord;
             SecurityInformation : SECURITY_INFORMATION;
             var StringSecurityDescriptor : LPTSTR;
             var StringSecurityDescriptorLen : ULONG
         ):boolean; stdcall;

{$EXTERNALSYM ConvertStringSecurityDescriptorToSecurityDescriptor}
function ConvertStringSecurityDescriptorToSecurityDescriptor(
             StringSecurityDescriptor : LPSTR;
             RequestedStringSDRevision : DWord;
             var SecurityDescriptor : PSECURITY_DESCRIPTOR;
             var SecurityDescriptorSize : ULONG
         ):boolean; stdcall;

{$EXTERNALSYM ConvertSidToStringSid}
function ConvertSidToStringSid(
             Sid:PSID;
             var StringSid : LPSTR
         ):boolean; stdcall;

implementation

uses SysUtils, ActiveX, ComObj, WbemScripting_TLB, Variants;

function ConvertSecurityDescriptorToStringSecurityDescriptor; external 'Advapi32.dll' name 'ConvertSecurityDescriptorToStringSecurityDescriptorA';
function ConvertStringSecurityDescriptorToSecurityDescriptor; external 'Advapi32.dll' name 'ConvertStringSecurityDescriptorToSecurityDescriptorA';
function ConvertSidToStringSid; external 'Advapi32.dll' name 'ConvertSidToStringSidA';

function addUser(user:string; a:olevariant):olevariant;
type TByteArray = array of byte;
var i: integer;
    sd: PSECURITY_DESCRIPTOR;
    ssd,
    ssid: PAnsiChar;
    ssdLen: dword;
    psa: PSafeArray;
    pd: pointer;
    tmp: string;

    arrayBounds: TSafeArrayBound;
    va: variant;

    sid:PSID;

    domain:PAnsiChar;
    peUse,
    domainSize,
    sidSize:cardinal;

    p:integer;
begin
  arrayBounds.lLbound:=0;
  arrayBounds.cElements:=VarArrayHighBound(a, 1);
  psa:=SafeArrayCreate(varByte, 1, arrayBounds);
  pd:=nil;
  if SafeArrayAccessData(psa, pd) = S_OK then
  begin
    for i:=0 to VarArrayHighBound(a, 1) do
      TByteArray(pd)[i]:=a[i];
    SafeArrayUnaccessData(psa);
  end;
                                
  sd:=PSECURITY_DESCRIPTOR(pd);

  if not IsValidSecurityDescriptor(sd) then
    raise Exception.Create('The security descriptor received is not valid! (1)');

  if not ConvertSecurityDescriptorToStringSecurityDescriptor(
                sd,
                1,//SDDL_REVISION_1,
                OWNER_SECURITY_INFORMATION or
                GROUP_SECURITY_INFORMATION or
                DACL_SECURITY_INFORMATION or
                SACL_SECURITY_INFORMATION
                ,
                ssd,
                ssdLen
              ) then
    raise Exception.Create('Could not covert security descriptor to a string security descriptor.');

  sidSize:=0;// first call = get necessary size
  domain:=nil;// first call = get necessary size
  sid:=AllocMem(1);
  LookupAccountName(nil, PAnsiChar(user), sid, sidSize, domain, domainSize, peUse);
  FreeMem(sid);
  sid:=AllocMem(sidSize);
  domain:=AllocMem(domainSize);
  LookupAccountName(nil, PAnsiChar(user), sid, sidSize, domain, domainSize, peUse);

  if not ConvertSidToStringSid(sid, ssid) then
    raise Exception.Create('Could not convert sid to string sid.');

  FreeMem(sid);
  FreeMem(domain);

  p:=pos(ssid,ssd)-1;// we are working with a pansichar which is 0-indexed
  if p>0 then
  begin
    i:=p;
    dec(i);
    dec(i);
    while ssd[i]<>';' do
      dec(i);
    dec(i);
    while ssd[i]<>';' do
      dec(i);
    p:=i;
    dec(i);
    while ssd[i]<>';' do
      dec(i);
    dec(i);
    while ssd[i]<>';' do
      dec(i);

    tmp:=copy(ssd,1,i+1)+
         'CI;CCDCLCSWRPWPRCWD'+ // set FULL access (including subnamespaces)
         copy(ssd,p+1,length(ssd));
  end              else
  begin
    tmp:=copy(ssd,1,pos('(',ssd)-1)+
         '(A;CI;CCDCLCSWRPWPRCWD;;;'+ // add FULL access (including subnamespaces)
         ssid+
         ')'+
         copy(ssd,pos('(',ssd),length(ssd));
  end;
  LocalFree(cardinal(ssid));
  LocalFree(cardinal(ssd));

  ssd:=PAnsiChar(tmp);
  
  if not ConvertStringSecurityDescriptorToSecurityDescriptor(
                ssd,
                1,//SDDL_REVISION_1,
                sd,
                ssdLen
              ) then
    raise Exception.Create('Could not convert string security descriptor to security descriptor.');

  if not IsValidSecurityDescriptor(sd) then
    raise Exception.Create('The converted security descriptor is not valid! (2)');

  va:=VarArrayCreate([0,ssdLen-1],varByte);
  for i:=0 to ssdLen-1 do
    va[i]:=TByteArray(sd)[i];
  LocalFree(cardinal(sd));
  result:=va;
end;{}

function setWMIAccessRights(user:string):string;
var
  SWbemLocator1:TSWbemLocator;
  Service: ISWbemServices;
  InParam,
  OutParam, SObject:             ISWbemObject;
  Method:                                 ISWbemMethod;
  SProp1
  , SProp2
  , MyProperty
  :             ISWbemProperty;
  s,methodName:string;
  PropValue:OleVariant;
begin
  methodName:='GetSD';
  SWbemLocator1:=TSWbemLocator.Create(nil);
  Service:=SWbemLocator1.ConnectServer('.','root', '', '', '','', 0, nil);
  Service.Security_.Set_ImpersonationLevel(wbemImpersonationLevelImpersonate);
  SObject:= Service.Get('__SystemSecurity', 0, nil);
  Method:= SOBject.Methods_.Item(methodName, 0);
  OutParam:= Method.OutParameters.SpawnInstance_(0);
  MyProperty:= OutParam.Properties_.Add('SD', wbemCimtypeUint8, true, 0);
//  MyProperty.Set_Value(PropValue);
  OutParam:= SObject.ExecMethod_(methodName, OutParam, 0, nil);
  SProp1:= outParam.Properties_.Item('ReturnValue', 0);
  case SProp1.Get_Value of
    S_OK:begin
           s:='OK';
           SProp2:= outParam.Properties_.Item('SD', 0);
         end;
//    WBEM_E_ACCESS_DENIED:s:='Caller does not have sufficient rights to call this method.';
//    WBEM_E_METHOD_DISABLED:s:='Attempted to run this method on a machine running Windows NT/Windows 2000/Windows XP.';
    else s:='execution returned unknown error value: '+SProp1.Get_Value;
  end;
  SWbemLocator1.Free;
  service:=nil;
  SObject:=nil;
  OutParam:=nil;
  SProp1:=nil;
  result:=s;
  if s<>'OK' then
    exit;

  methodName:='SetSD';
  SWbemLocator1:=TSWbemLocator.Create(nil);
  Service:=SWbemLocator1.ConnectServer('.','root', '', '', '','', 0, nil);
//  Service.Security_.Set_ImpersonationLevel(wbemImpersonationLevelImpersonate);
  SObject:= Service.Get('__SystemSecurity', 0, nil);
  Method:= SOBject.Methods_.Item(methodName, 0);
  InParam:= Method.InParameters.SpawnInstance_(0);
  MyProperty:= InParam.Properties_.Add('SD', wbemCimtypeUint8, true, 0);
  PropValue:=SProp2.Get_Value;

  propvalue:=addUser(user, propvalue);// add user credentials to SD

  MyProperty.Set_Value(PropValue);
  OutParam:= SObject.ExecMethod_(methodName, InParam, 0, nil);
  SProp1:= outParam.Properties_.Item('ReturnValue', 0);
  case SProp1.Get_Value of
    S_OK:begin
           s:='OK';
         end;
//    WBEM_E_ACCESS_DENIED:s:='Caller does not have sufficient rights to call this method.';
//    WBEM_E_METHOD_DISABLED:s:='Attempted to run this method on a machine running Windows NT/Windows 2000/Windows XP.';
    else s:='execution returned unknown error value: '+SProp1.Get_Value;
  end;
  SWbemLocator1.Free;
  service:=nil;
  SObject:=nil;
  OutParam:=nil;
  SProp1:=nil;
end;

function IsAdmin: Boolean;
var
  hAccessToken: THandle;
  ptgGroups: PTokenGroups;
  dwInfoBufferSize: DWORD;
  psidAdministrators: PSID;
  x: Integer;
  bSuccess: BOOL;
begin
  Result   := False;
  bSuccess := OpenThreadToken(GetCurrentThread, TOKEN_QUERY, True,
    hAccessToken);
  if not bSuccess then
  begin
    if GetLastError = ERROR_NO_TOKEN then
      bSuccess := OpenProcessToken(GetCurrentProcess, TOKEN_QUERY,
        hAccessToken);
  end;
  if bSuccess then
  begin
    GetMem(ptgGroups, 1024);
    bSuccess := GetTokenInformation(hAccessToken, TokenGroups,
      ptgGroups, 1024, dwInfoBufferSize);
    CloseHandle(hAccessToken);
    if bSuccess then
    begin
      AllocateAndInitializeSid(SECURITY_NT_AUTHORITY, 2,
        SECURITY_BUILTIN_DOMAIN_RID, DOMAIN_ALIAS_RID_ADMINS,
        0, 0, 0, 0, 0, 0, psidAdministrators);
      {$R-}
      for x := 0 to ptgGroups.GroupCount - 1 do
        if EqualSid(psidAdministrators, ptgGroups.Groups[x].Sid) then
        begin
          Result := True;
          Break;
        end;
      {$R+}
      FreeSid(psidAdministrators);
    end;
    FreeMem(ptgGroups);
  end;
end;

end.
